import threading

from requests_throttler.utils import locked


class ThrottledRequestAlreadyFinished(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class ThrottledRequest(object):
    """This class represents a throttled request


    Attributes:
        `_request` - the prepared request to throttle
        `_finished` - the flag that indicates if the request has been sent and a response has
                      been received or an exception occured
        `_response` - the response corresponding to the request
        `_exception` - the exception occured during the request (`None` if no exceptions
                       occured)
        `not_done` - the condition on which to wait to have the response an to make the
                     object thread-safe

    """

    def __init__(self, request):
        """Create a throttled request with the given prepared request

        `request` - the prepared request to throttle

        """
        self._request = request
        self._finished = False
        self._response = None
        self._exception = None
        self.not_done = threading.Condition(threading.Lock())

    def __str__(self):
        return "[{class_name} <{request}, {response}, {finished}, {exception}>]".format(
            class_name="ThrottledRequest",
            request=repr(self._request),
            response=repr(self._response),
            finished=repr(self._finished),
            exception=repr(self._exception))

    @property
    @locked('not_done')
    def request(self):
        """Return the corresponding request"""

        return self._request

    @property
    @locked('not_done')
    def finished(self):
        """Return the flag that indicates if the request has been processed"""

        return self._finished

    @property
    def response(self):
        """Return the response obtained by processing the request (blocking)"""

        return self.get_response(timeout=None)

    @response.setter
    @locked('not_done')
    def response(self, response):
        """Set the response that has been received during the processing of the request

        It raises `ThrottledRequestAlreadyFinshed` if the throttled request has already
        finished.

        """
        if self._finished is True:
            raise ThrottledRequestAlreadyFinished("ThrottledRequest already finished")
        self._response = response
        self._finished = True
        self.not_done.notify()

    @locked('not_done')
    def get_response(self, timeout=0):
        """Return the response obtained by processing the request

        If `timeout` is `None` it waits until a response has been obtained or an execption
        occurs. When a response has been obtained it is returned, if an exception occurs it
        is raised.
        If `timeout` is not `None` and no response still hasn't been obtained `None` is
        returned.

        `timeout` - the timeout value in seconds (default: 0)

        """
        if self._wait_finished(timeout):
            if self._exception is None:
                return self._response
            raise self._exception
        else:
            return None

    @property
    def exception(self):
        """Return the exception that occurs by processing the request (blocking)"""

        return self.get_exception(timeout=None)

    @exception.setter
    @locked('not_done')
    def exception(self, exception):
        """Set the exception that has been raised during the processing of the request"""

        if self._finished is True:
            raise ThrottledRequestAlreadyFinished("ThrottledRequest already finished.")

        self._exception = exception
        self._finished = True
        self.not_done.notify()

    @locked('not_done')
    def get_exception(self, timeout=0):
        """Return the exception that occurs by processing the request

        If `timeout` is `None` it waits until a response has been obtained or an execption
        occurs. If an exception occurs it is returned otherwise `None` is returned.
        If `timeout` is not `None` and no response still hasn't been obtained `None` is
        returned.

        `timeout` - the timeout value in seconds (default: 0)

        """
        if self._wait_finished(timeout):
            return self._exception
        return None

    def _wait_finished(self, timeout=None):
        """Wait for the request for being finished with an optional `timeout`

        The waiting is blocking and can last indefenetely if `timeout` is `None`. Returns
        `True` if the request has been processed and `False` otherwise.

        `timeout` - the timeout value in seconds (default: 0)

        """
        if timeout is None:
            while not self._finished:
                self.not_done.wait()
        else:
            self.not_done.wait(timeout)
            if not self._finished:
                return False
        return True



import time


class NoCheckpointSetError(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Timer(object):
    """This class provides an utility to calculates elapsed time since a start/check point


    Attributes:
        `_start` - the starting time of the timer
        `_checkpoint` - the last checkpoint of the timer

    """

    def __init__(self, start=None, checkpoint=None):
        """Create the timer with the given starting time and an eventual checkpoint

        `start` - the starting time of the timer (default: now)
        `checkpoint` - the checkpoint (default: `None`)

        """
        self._start = start if start is not None else time.time()
        self._checkpoint = checkpoint

    @property
    def start(self):
        """Return the timer starting time"""

        return self._start

    @property
    def checkpoint(self):
        """Return the last checkpoint"""

        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, checkpoint):
        """Set a checkpoint"""

        self._checkpoint = checkpoint

    def total_elapsed(self):
        """Return the elapsed time since the timer starting time"""

        return time.time() - self._start

    def elapsed(self):
        """Return the elapsed time since the last checkpoint

        It raises `NoCheckpointSetError` if `checkpoint` is `None`.

        """
        if self._checkpoint is None:
            raise NoCheckpointSetError("No checkpoint has been set.")
        return time.time() - self._checkpoint

    def get_elapsed_and_set_checkpoint(self, change=True, new_checkpoint=None):
        """Return the elapsed time since the last checkpoint and change the checkpoint

        The checkpoint is changed with `new_checkpoint` if `change` is `True`. If
        `new_checkpoint` is `None`, then to checkpoint is assigned the current time. It
        raises `NoCheckpointSetError` if `checkpoint` is `None`.

        `change` - the flag that indicates if the checkpoint is to be changed (default: `True`)
        `new_checkpoint` - the new checkpoint to assign (default: `None`)

        """
        if self._checkpoint is None:
            raise NoCheckpointSetError("No checkpoint has been set.")
        now = time.time()
        elapsed = now - self._checkpoint
        if change:
            self._checkpoint = now if new_checkpoint is None else new_checkpoint
        return elapsed



import time
import threading
from collections import deque as queue
from concurrent.futures import ThreadPoolExecutor

import requests

from requests_throttler.utils.timer import Timer
from requests_throttler.utils import locked, get_logger
from requests_throttler.throttled_request.throttled_request import ThrottledRequest

logger = get_logger(__name__)

THROTTLER_STATUS = ['initialized', 'running', 'waiting', 'paused', 'stopped', 'ending', 'ended']
THROTTLER_STATUS_DEPENDENCIES = {'initialized': ['initialized', 'running', 'stopped'],
                                 'running': ['running', 'waiting', 'paused', 'stopped'],
                                 'waiting': ['waiting', 'running', 'paused', 'stopped'],
                                 'paused': ['paused', 'running', 'waiting', 'stopped'],
                                 'stopped': ['stopped', 'ending'],
                                 'ending': ['ending', 'ended']}


class ThrottlerStatusError(Exception):
    
    def __init__(self, msg, current_status, previous_status=None):
        self.msg = msg
        self.status = current_status
        self.previous_status = previous_status

    def __str__(self):
        if self.previous_status is None:
            return "{message} (current status: {status})".format(message=self.msg,
                                                                 status=self.status)
        return "{message} ({prev_status} ---> {status})".format(message=self.msg,
                                                                status=self.status,
                                                                prev_status=self.previous_status)


class FullRequestsPoolError(Exception):
    
    def __init__(self, msg, pool):
        self.msg = msg
        self.pool = pool

    def __str__(self):
        return "{message} (pool size: {pool_size})".format(message=self.msg,
                                                           pool_size=self.pool.maxlen)


class BaseThrottler(object):
    """This class provides the base requests throttler

    The base throttler guarantees that between each request a fixed amount of time between them
    elapsed. The pool can be limited by a maximum length and an exception is raised if a request
    is tried to be enqueued in the full pool.


    Attributes:
        `_name` - the name of the throttler
        `_requests_pool` - the pool containing the requests (FIFO)
        `_delay` - the delay in seconds between each request
        `_status` - the current status of the thottler
        `_session` - the session to perform the requests
        `_executor` - the executor responsable to start the throttler
        `_timer` - the timer responsable to measure the time between each request
        `_successes` - the number of request that succeded
        `_failures` - the number of request that failed
        `_wait_enqueued` - a flag that indicates if after the shutdown the requests enqueued
                           have to be finished or aborted
        `status_lock` - the condition on which to wait on a specific status change
        `not_empty` - the condition on which to wait when the pool of requests is empty

    """

    def __init__(self, name=None, delay=0, max_pool_size=None):
        """Create a base throttler with the given delay time and pool size

        If `delay` is a negative number then `ValueError` is raised.

        `name` - the name of the throttler (default: `None`)
        `delay` - the fixed positive amount of time that must elapsed bewteen each request
                  in seconds (default: 0)
        `max_pool_size` - the maximum number of enqueueable requests (default: unlimited)

        """
        if delay < 0:
            raise ValueError("The delay value must be positive.")
        self._name = name
        self._requests_pool = queue(maxlen=max_pool_size)
        self._delay = delay
        self._status = 'initialized'
        self._session = requests.Session()
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._timer = Timer(checkpoint=0)
        self._successes = 0
        self._failures = 0
        self._wait_enqueued = None
        self.status_lock = threading.Condition(threading.Lock())
        self.not_empty = threading.Condition(threading.Lock())

    def __str__(self):
        return "[{class_name} <{name}, {delay}, {status}>]".format(class_name="BaseThrottler",
                                                                   name=repr(self._name),
                                                                   delay=repr(self._delay),
                                                                   status=repr(self._status))

    def __enter__(self):
        """Start the throttler by entering in a context"""

        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Shutdown the throttler by exiting from the context"""

        self.shutdown()
        return value

    @property
    def name(self):
        """Return the name of the throttler"""

        return self._name

    @property
    def delay(self):
        """Return the delay value between each request"""

        return self._delay

    @property
    @locked('status_lock')
    def status(self):
        """Return the status of the throttler"""

        return self._status

    @status.setter
    @locked('status_lock')
    def status(self, status):
        """Set a new status

        If `status` is invalid then `ThrottlerStatusError` is raised.

        `status` - the new status to assign

        """
        if status not in THROTTLER_STATUS:
            raise ThrottlerStatusError("Invalid status.", status)
        if status not in THROTTLER_STATUS_DEPENDENCIES[self._status]:
            raise ThrottlerStatusError("Invalid status stransition.", status,
                                       previous_status=self._status)
        logger.debug("Status changing: %s ---> %s", self._status, status)
        self._status = status

    @property
    @locked('status_lock')
    def successes(self):
        """Return the number of successes"""

        return self._successes

    @property
    @locked('status_lock')
    def failures(self):
        """Return the number of failures"""

        return self._failures

    def start(self):
        """Start the throttler by starting the main loop

        If the throller has been already started then `ThrottlerStatusError` is raised.

        """
        logger.info("Starting base throttler '%s'...", self._name)
        if self.status in ['running', 'waiting', 'paused']:
            raise ThrottlerStatusError("Cannot start an already started throttler.",
                                       self._status)
        if self.status in ['stopped', 'ending', 'ended']:
            raise ThrottlerStatusError("Cannot start an already shutdown throttler.",
                                       self._status)

        self.status = 'running'
        self._executor.submit(self._main_loop)

    @locked('not_empty')
    def shutdown(self, wait_enqueued=True):
        """Shutdown the throttler by shutdowning the executor

        If `wait_enqueued` is `True` then before stopping the throttlers consumes all requests
        enqueued. Otherwise the throttler is forced to be shutdowned

        `wait_enqueued` - the flag that indicates if the already enqueued requests are to be
                          processed or aborted

        """
        if self.status in ['stopped', 'ending', 'ended']:
            raise ThrottlerStatusError("Cannot shutdown an already shutdown throttler.",
                                       self._status)
        self.status = 'stopped'
        self._wait_enqueued = wait_enqueued
        self.not_empty.notify()
        self._executor.shutdown(wait=False)

    @locked('status_lock')
    def pause(self):
        """Pause the throttler

        If the throttler is not `running` or `waiting` then `ThrottlerStatusError` is raised.

        """
        if self._status == 'paused':
            raise ThrottlerStatusError("Cannot pause an already paused throttler", self._status)
        if self._status not in ['running', 'waiting']:
            raise ThrottlerStatusError("Cannot pause a not running throttler", self._status)

        self._status = 'paused'
        self.status_lock.notify()

    @locked('status_lock')
    def unpause(self):
        """Unpauses the throttler

        If the throttler is not `paused` then `ThrottlerStatusError` is raised.

        """
        if self._status != 'paused':
            raise ThrottlerStatusError("Cannot unpause not paused throttler", self._status)
        self._status = 'running'
        self.status_lock.notify()

    def submit(self, reqs):
        """Submit a single request or an entire list and return the corresponding response

        If `reqs` is a single request then the corresponding throttled request is returned,
        otherwise if `reqs` is a `list` then the list of corresponding throttled requests
        is returned.

        `reqs` - it can be a single request or a list of them
        
        """
        if isinstance(reqs, list):
            return [self._submit(r) for r in reqs]
        return self._submit(reqs)

    def _submit(self, request):
        """Submits the given request by preparing it and enqueueing it

        If the throttler is not `running`, `paused` or `waiting` an exception is raised.
        With the returned throttled request is associated an eventual exception that could be
        raised during the preparation or the enqueueing.

        `request` - the request to submit

        """
        logger.info("Submitting request to base throttler (url: %s)...", request.url)
        if self._status not in ['running', 'paused', 'waiting']:
            raise ThrottlerStatusError("Cannot submit request to throttler", self._status)
        throttled_request, prepared = self._prepare_request(request)
        if prepared:
            try:
                self._enqueue_request(throttled_request)
            except FullRequestsPoolError as e:
                throttled_request.exception(e)
                self._inc_failures()
        return throttled_request

    def _main_loop(self):
        """The main loop of the throttler"""

        logger.info("Starting main loop...")
        while True:
            self._sleep_or_pause()
            next_request = self._dequeue_request()
            if next_request is None:
                break
            self._send_request(next_request)
        logger.info("Exited from main loop.")
        self._end()

    @locked('status_lock')
    def _end(self):
        """Set the `ended` status"""

        self._status = 'ended'
        self.status_lock.notify()

    @locked('status_lock')
    def wait_end(self):
        """Wait untile the throttler is `ended`"""

        while self._status != 'ended':
            self.status_lock.wait()

    @locked('status_lock')
    def _sleep_or_pause(self):
        """Sleep or pause depending on the status"""

        while self._status == 'paused':
            logger.info("Pausing...")
            self.status_lock.wait()
            logger.info("Unpaused!")
        if not self._status == 'stopped':
            remaining_time = self._remaining_time()
            if remaining_time > 0:
                logger.debug("Start sleeping for %f seconds...", remaining_time)
                time.sleep(remaining_time)
                logger.debug("Awakening...")
            self._timer.checkpoint = time.time()

    def _remaining_time(self):
        """Return the remaining time before performing the next request"""

        return self._delay - self._timer.elapsed()

    def _prepare_request(self, request):
        """Prepare the given request and return the corresponding throttled request

        If an exception occurs during the preparation it is associated to the throttled request
        created.

        `request` - the request to prepare

        """
        try:
            logger.debug("Preparing request (url: %s)...", request.url)
            prepared_request = request.prepare()
        except requests.exceptions.RequestException as e:
            throttled_request = ThrottledRequest(None)
            throttled_request.exception = e
            self._inc_failures()
            prepared = False
            logger.warning("Unable to prepare the request (url: %s).", request.url)
        else:
            throttled_request = ThrottledRequest(prepared_request)
            prepared = True
            logger.debug("Request prepared!")
        return throttled_request, prepared

    def _send_request(self, throttled_request):
        """Send the given throttled request

        If an exception occurs during the sending it is associated to the throttled request.

        `throttled_request` - the throttled request to send
        
        """
        try:
            logger.info("Sending request (url: %s)...", throttled_request.request.url)
            response = self._session.send(throttled_request.request)
        except Exception as e:
            throttled_request.exception = e
            self._inc_failures()
            logger.warning("Unable to send the request (url: %s).",
                           throttled_request.request.url)
        else:
            throttled_request.response = response
            self._inc_successes()
            logger.info("Request sent! (url: %s)", throttled_request.request.url)

    @locked('not_empty')
    def _enqueue_request(self, throttled_request):
        """Enqueue the given throttled request

        If the the pool of requests is full `FullRequestPoolError` is raised.

        `throttled_request` - the throttled request to enqueue

        """
        logger.debug("Enqueueing request (url: %s)...", throttled_request.request.url)
        if len(self._requests_pool) == self._requests_pool.maxlen:
            raise FullRequestsPoolError("The requests pool is full.", self._requests_pool)

        self._requests_pool.append(throttled_request)
        logger.debug("Request enqueued! (url: %s)", throttled_request.request.url)
        self.not_empty.notify()

    @locked('not_empty')
    def _dequeue_request(self):
        """Dequeue the next throttled request to process and return it

        If the throttler is `running` and no requests are eunqueued the throttler waits until
        a new request arrives.

        """
        logger.debug("Dequeueing request...")
        waiting = True
        while waiting:
            waiting, proceed = self._dequeue_condition()
            if waiting:
                logger.info("Start waiting for new requests...")
                self.not_empty.wait()
                logger.info("Awakening...")
            else:
                waiting = False
        
        if proceed:
            next_request = self._requests_pool.popleft()
        else:
            return None
        logger.debug("Request dequeued! (url: %s)", next_request.request.url)
        self.status = 'running' if self.status not in ['stopped', 'ending'] else self.status
        return next_request

    def _dequeue_condition(self):
        """Check if the throttler has to wait or has to proceed

        Returns a tuple of the form (`waiting`, `proceed`) where `waiting` indicates if the
        throttler has to wait while `proceed` indicates if the throttler has to proceed.

        """
        if self.status == 'stopped':
            self.status = 'ending'
            if not self._wait_enqueued:
                return False, False

        if self.status == 'ending':
            if len(self._requests_pool) == 0:
                return False, False

        if self.status == 'paused':
            return True, False

        if len(self._requests_pool) == 0:
            self.status = 'waiting'
            return True, False

        return False, True

    @locked('status_lock')
    def _inc_successes(self):
        """Increment the number of successes"""

        self._successes += 1

    @locked('status_lock')
    def _inc_failures(self):
        """Increment the number of failures"""

        self._failures += 1



import logging
import threading

from requests_throttler.settings import \
    LOG_FORMAT, \
    DEFAULT_LOG_LEVEL


def locked(lock):
    """Decorator usefull to access to a function with a lock named `lock`"""

    def _locked(func):
        def wrapper(*args, **kwargs):
            lock_to_use = getattr(args[0], lock)
            if not isinstance(lock_to_use, threading.Lock().__class__) and \
               not isinstance(lock_to_use, threading.Condition().__class__):
                raise ValueError('`{lock}` is not an instance neither of threading.Lock neither '
                                 'of threading.Condition'.format(lock=lock_to_use))
            with lock_to_use:
                return func(*args, **kwargs)

        return wrapper
    return _locked


def get_logger(name, level=DEFAULT_LOG_LEVEL):
    logging.basicConfig(format=LOG_FORMAT[level])
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger



from setuptools import setup

import requests_throttler


packages = ['requests_throttler',
            'requests_throttler.throttled_request',
            'requests_throttler.throttler',
            'requests_throttler.utils',
            'requests_throttler.tests']

requires = ['requests==2.1.0',
            'futures==2.1.5']

classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Topic :: Internet',
    'Topic :: Internet :: WWW/HTTP'
]

with open('README.rst') as f:
    readme = f.read()
with open('HISTORY.rst') as f:
    history = f.read()

setup(name=requests_throttler.__title__,
      version=requests_throttler.__version__,
      license=requests_throttler.__license__,
      url=requests_throttler.__project_url__,
      description='Python HTTP requests throttler',
      long_description=readme + '\n\n' + history,
      author=requests_throttler.__author__,
      author_email=requests_throttler.__author_email__,
      packages=packages,
      package_dir={'requests_throttler': 'requests_throttler'},
      package_data={'': ['LICENSE', 'NOTICE']},
      include_package_data=True,
      install_requires=requires,
      classifiers=classifiers)
